/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "dq-module.hh"
#include "DVecType.hh"
#include "Bits.hh"
#include "checksum_crc32.hh"
#include "jsonStack.hh"
//------------------------------ filter header files.
#include "FilterDesign.hh"
#include "FIRFilter.hh"
#include "MultiPipe.hh"
#include "autopipe.hh"
#include "GateGen.hh"
#include "YSynch.hh"
#include "logic_2op.hh"
#include "math_2op.hh"
#include "MathOp.hh"
//------------------------------ system headers
#include <sstream>
#include <fstream>
#include <iostream>
#include <stdexcept>
#include <algorithm>
#include <dlfcn.h>

using namespace std;

EXECDAT(DQ_Module)

static TSeries null_ts;

DQ_Module::DQ_Module(int argc, const char* argv[])
 : MultiStream(argc, argv), stride(4.0), flag_step(0.0625),
   parity_bit(0), invert(false), config_version("Not Specified"), 
   config_crc32(0), frame_path("$HOME/frames/X-DQ_MODULE-%05r", "X-DQ_MODULE"),
   compression("uncompressed"), output_start(0), current(0), frames_out(0)
{
   try {
      config_file = frame_name::resolve_env(argv[argc - 1]);
      config_file = frame_name::make_absolute(config_file);
      config(config_file);
   } catch (std::exception& e) {
      cerr << "Caught exception reading configuration file: " 
	   << e.what() << endl;
      finish();
      return;
   }
   getDacc().setStride(stride);
   setStrideAlignment(stride, 0.0);
   init_output_channels();
   book_channels();
   fr_writer.setCompress(FrVectRef::get_compression_mode(compression));
   if (Debug()) put_config(cout);
}


//======================================  Destroy monitor object.
DQ_Module::~DQ_Module(void) {
   //-----------------------------------  write buffered output
   flush();
}

//======================================  Flush results of buffered inputs 
void 
DQ_Module::flush(void) {
   if (!current) return;

   //----------------------------------- loop over processed channels, flush
   for (auto itr=proc_list.begin(); itr!=proc_list.end(); itr++) {
      const TSeries& X(find(itr->channel_x()));
      Time Xt0 = X.getStartTime();
      TSeries Xend(X.extract(Xt0, current - Xt0));
      if (itr->channel_y().empty()) {
   	 if (!Xend.empty()) itr->apply(Xend);
      } else {
   	 const TSeries& Y(find(itr->channel_y()));
	 Time Yt0 = Y.getStartTime();
	 TSeries Yend(Y.extract(Yt0, current - Yt0));
	 if (!Xend.empty() || !Yend.empty()) itr->apply(X, Y);
      }
      itr->flush();
   }

   //-----------------------------------  Buffer for writing
   Time t0(0), tEnd(0);
   for (auto itr=proc_list.begin(); itr != proc_list.end(); itr++) {
      const TSeries& tsi = itr->buffer();
      if (!itr->write() || tsi.empty()) {
	 continue;
      }
      if (!t0) {
	 t0   = tsi.getStartTime();
	 tEnd = tsi.getEndTime();
      } else {
	 if (t0   < tsi.getStartTime())  t0 = tsi.getStartTime();
	 if (tEnd > tsi.getEndTime())  tEnd = tsi.getEndTime();
      }
   }
   if (Debug()) cout << "Available data: " << t0 << " - " << tEnd << endl;
   
   if (t0 < write_time) align_buffer(write_time);
   
   //-----------------------------------  Write the last frames.
   Interval dT = stride;
   while (write_time + dT <= tEnd) {
      write_frame(write_time, dT);
      write_time += dT;
   }
   if (tEnd > write_time) {
      dT = tEnd - write_time;
      write_frame(write_time, dT);
      write_time = tEnd;
   }
}

//======================================  Convert bool to char*
const char*
boolstr(bool torf) {
   return torf ? "true" : "false";
}

//======================================  Print configuration info

void 
DQ_Module::put_config(ostream& out) const {
   out << "Monitor configuration" << endl;
   out << "config-file:    " << config_file << endl;
   out << "config-version: " << config_version << endl;
   out << "config-crc32:   " << config_crc32 << endl;
   out << "stride:         " << stride << endl;
   out << "parity-bit      ";
   if (parity_bit < 0) out << "off" << endl;
   else                out << parity_bit << endl;
   out << "invert          " << boolstr(invert) << endl;
   out << "flag_step:      " << flag_step << endl;
   out << "Frame_name:     " << frame_path.file_path(Time(0)) << endl;
   out << "compression:    " << compression << endl;
   out << "DQ configuration" << endl;
   size_t nFlags = flag_vect.size();
   for (size_t i=0; i<nFlags; i++) {
      flag_vect[i]->put_config(out);
      out << endl;
   }
   
   out << "DQ processed channels" << endl;
   for (auto itr=proc_list.begin(); itr != proc_list.end(); itr++) {
      itr->dump(out);
      out << endl;
   }
   
   out << "DQ copy history" << endl;
   size_t nHist = copy_history.size();
   for (size_t i=0; i<nHist; i++) {
      out << "                " << copy_history[i] << endl;
   }
}

//=======================================  Parse json configuration
void 
DQ_Module::config(const std::string& cfg_file) {
   const char* valid_names[] =
      {"config-version", "stride", "flag_step", "parity-bit", "invert",
       "write_path", "copy_channels", "compression", "copy_history",
       "dq_flags", "filters", "processed-channels", 0};
	 
   //------------------------------------  Read the configuration file.
   ifstream in(cfg_file.c_str());
   if (!in.is_open()) {
      throw runtime_error(string("unable to open file: ") + cfg_file);
   }
   string conf_str;
   getline(in, conf_str, (char)EOF);
   if (conf_str.empty()) {
      throw runtime_error(string("Configuration file: ") 
	                  + cfg_file + " is empty");
   }

   //------------------------------------  Immortalize the file crc32 checksum
   checksum_crc32 crc;
   crc.add(conf_str);
   config_crc32 = crc.result();

   //------------------------------------  Parse out the file.
   bool syntax = false;
   jsonStack jstak(cfg_file);

   //------------------------------------  Check that everything is defined
   string bad = jstak.validate(valid_names);
   if (!bad.empty()) {
      cout << "dq-module: *** Warning *** invalid structure menber names in "
	 << cfg_file << " as follows: " << bad << endl;
   }				  

   jstak.fetch_data("config-version", config_version);
   jstak.fetch_data("stride", stride);
   jstak.fetch_data("flag_step", flag_step);
   if (jstak.push_member("parity-bit")) {
      jsonStack::autopop pop_pb(jstak);
      if (jstak.isBool()) {
	 bool pb = false;
	 jstak.fetch_value(pb);
	 if (!pb) parity_bit = -1;
      } else if (!jstak.fetch_value(parity_bit)) {
	 cerr << "Invalid parity-bit specification" << endl;
	 syntax = true;
      }
      if (parity_bit >= 32) {
	 cerr << "invalid parity bit number" << endl;
	 syntax = true;
      }
   }
   jstak.fetch_data("invert", invert);
   string write_path;
   if (jstak.fetch_data("write_path", write_path)) frame_path.split(write_path);

   vector<string> copy_channels;
   if (jstak.fetch_data("copy_channels", copy_channels)) {
      for (auto itr=copy_channels.begin(); itr!=copy_channels.end(); itr++) {
	 string chan = *itr;
	 size_t inx = chan.find("/");
	 if (inx != string::npos) {
	    proc_list.push_back(process_channel(chan.substr(inx+1)));
	 } else {
	    proc_list.push_back(process_channel(chan));
	 }
	 proc_list.back().set_channel_x(chan);
	 proc_list.back().set_process_type(process_channel::pk_copy);
      }
   }
   
   jstak.fetch_data("compression", compression);

   jstak.fetch_data("copy_history", copy_history);

   //---------------------------------- Look for dq_flags array
   if (!jstak.push_member("dq_flags")) {
      cerr << "No flags defined in document: " << cfg_file << endl;
   }
   //---------------------------------- Parse the dq_flags structure
   else {
      jsonStack::autopop pop_flags(jstak);
      if (Debug()) cout << "config: loop over dq_flags" << endl;
      for (int i=0; jstak.iterate(); i++) {
	 jsonStack::autopop pop_iter(jstak);
	 string name, chan;
	 jstak.fetch_data("name", name);

	 if (!jstak.fetch_data("channel", chan)) {
	    cerr << "DQ_Module::config Flag " << name
		 << ", channel is not specified" << endl;
	    syntax = true;
	    continue;
	 }
	 if (Debug()) cout << "config: flag: " << i << " name: " << name
			   << " channel: " << chan << endl;

	 int bit = 0;
	 if (!jstak.fetch_data("bit", bit)) {
	    cerr << "DQ_Module::config Flag " << name
		 << ", bit number is not specified" << endl;
	    syntax = true;
	 }
      
	 if (name.empty()) {
	    ostringstream pseudo;
	    pseudo << chan << "[" << bit << "]";
	    name = pseudo.str();
	    cout << "Unnamed flag referred to as " << name << endl; 
	 }

	 string plugin;
	 if (!jstak.fetch_data("class", plugin)) {
	    cerr << "DQ_Module::config: Flag " << name
		 << ", class name is not specified" << endl;
	    syntax = true;
	 }

	 string library = "/usr/lib/libdq-module.so";
	 if (!jstak.fetch_data("library", library)) {
	    cerr << "DQ_Module::config: Flag " << name
		 << ", library is not specified." << library << " is assumed."
		 << endl;
	    syntax = true;
	 }

	 //-----------------------------  Add channel list
	 DQ_bit::chanlist_type chan_list;
	 if (!jstak.fetch_data("inputs", chan_list)) {
	    cerr << "No input channels specified for DQ Flag " << name << endl;
	    syntax = true;
	 }
	 
	 //-----------------------------  Make executable symbols global
	 void* handle = dlopen("",  RTLD_NOW | RTLD_GLOBAL);
	 if (handle == 0) {
	    perror("dlopen error");
	    cerr << "Unable to load executable." << endl;
	    cerr << dlerror() << endl;
	    syntax = true;
	    continue;
	 }

	 //-----------------------------  Load the plugin
         handle = dlopen (library.c_str(), RTLD_NOW | RTLD_GLOBAL);
	 if (handle == 0) {
	    perror("dlopen error");
	    cerr << "Unable to load " << library << endl;
	    cerr << dlerror() << endl;
	    syntax = true;
	    continue;
	 }

	 typedef DQ_bit* (*func_t) (void);
	 string func = plugin + "_init";
	 func_t plug_init=reinterpret_cast<func_t>(dlsym(handle, func.c_str()));
	 if (plug_init == 0) {
	    cerr << "dlsym failed for " << func << endl;
	    syntax = true;
	    continue;
	 }

	 //----------------------------  Construct the channel
	 DQ_bit* dq_obj = plug_init();
	 dq_obj->dq_class(plugin);
	 dq_obj->dq_name(name);
	 dq_obj->bit_channel(chan);
	 dq_obj->bit_number(bit);
	 dq_obj->channel_list(chan_list);

	 //-----------------------------  Add parameter lists
	 if (jstak.push_member("parameters")) {
	    jsonStack::autopop pop_params(jstak);
	    jsonStack::string_vect pars = jstak.getNames();
	    size_t nPar = pars.size();
	    for (size_t j=0; j < nPar; j++) {
	       string par_name = pars[j];
	       jstak.push_member(par_name);
	       if (jstak.isNumeric()) {
		  double val = 0;
		  jstak.fetch_value(val);
		  dq_obj->numeric_param(par_name, val);
	       }
	       else if (jstak.isBool()) {
		  bool val = false;
		  jstak.fetch_value(val);
		  double dval = val ? 1.0 : 0.0;
		  dq_obj->numeric_param(par_name, dval);
	       }
	       else {
		  string val;
		  jstak.fetch_value(val);
		  dq_obj->string_param(par_name, val);
	       }
	       jstak.pop();
	    }
	 }

	 //-----------------------------  Check the invert bit
	 bool inv_flag = false;
	 if (jstak.fetch_data("invert", inv_flag)) {
	    dq_obj->invert(inv_flag);
	 }
	 
	 //-----------------------------  Add this flag to the list
	 flag_vect.push_back(dq_obj);
      }
   }
   
   //------------------------------------  build filter definitions
   if (jstak.push_member("filters")) {
      jsonStack::autopop pop_flags(jstak);
      if (Debug()) cout << "config: loop over filters" << endl;
      for (int i=0; jstak.iterate(); i++) {
	 jsonStack::autopop pop_iter(jstak);
	 string name, type;
	 filter_template fdtemp;
	 if (!jstak.fetch_data("name", name) ||
	     !jstak.fetch_data("type", type)) {
	    syntax = true;
	    cerr << "filter definition requires name and type fields" << endl;
	 }

	 //-----------------------------  FilterDesign
	 else if (type == "design") {
	    string design;
	    double rate = 0;
	    if (!jstak.fetch_data("design", design) ||
		!jstak.fetch_data("rate", rate)) {
	       cerr << "filterdesign filters require 'design' and 'rate'"
		    << endl;
	       syntax = true;
	       continue;
	    }
	    FilterDesign fd(design.c_str(), rate);
	    fd.setFirType(6);
	    fdtemp = filter_template(name, fd.get(), filter_template::kf_pipe);
	 }

	 //-----------------------------  GateGen
	 else if (type == "GateGen") {
	    double rate(16384.), front(0.25), transit(0.25), minwidth(-1);
	    double threshold(1), idle(0), active(1);
	    string window("tukey"), oper(">");
	    if (!jstak.fetch_data("rate", rate)) {
	       cerr << "GateGen: Sample rate argument is required" << endl;
	       syntax = true;
	       continue;
	    }
	    jstak.fetch_data("front", front);
	    jstak.fetch_data("transition", transit);
	    jstak.fetch_data("min-width", minwidth);
	    jstak.fetch_data("window", window);
	    GateGen gg(rate, transit, front, minwidth, window);
	    bool setsel =   jstak.fetch_data("threshold", threshold)
	                  | jstak.fetch_data("bit-mask", threshold);
	    if (jstak.fetch_data("mode", oper) || setsel) {
	       gg.set_selection(oper, threshold);
	    }

	    jstak.fetch_data("idle",   idle);
	    jstak.fetch_data("active", active);
	    gg.set_idle(idle);
	    gg.set_active(active);

	    fdtemp = filter_template(name, gg, filter_template::kf_pipe);
	    if (Debug()) gg.dump(cout);
	 }

	 //-----------------------------  mathop
	 else if (type == "MathOp") {
	    string opcode("*");
	    double value(1);
	    jstak.fetch_data("opcode", opcode);
	    jstak.fetch_data("value",  value);
	    MathOp mop(opcode, value);
	    fdtemp = filter_template(name, mop, filter_template::kf_pipe);
	 }

	 //-----------------------------  logic_2op
	 else if (type == "logic_2op") {
	    string oper("A|B");
	    jstak.fetch_data("opcode", oper);
	    fdtemp = filter_template(name, YSynch(logic_2op(oper)),
				     filter_template::kf_ypipe);	       
	 }

	 //-----------------------------  math_2op
	 else if (type == "math_2op") {	       
	    string oper("+");
	    jstak.fetch_data("opcode", oper);
	    fdtemp = filter_template(name, YSynch(math_2op(oper)),
				     filter_template::kf_ypipe);	       
	 }

	 //-----------------------------  Unknown/undefined filter type
	 else {
	    cerr << "Unknown filter type: " << type << endl;
	    syntax = true;
	    continue;
	 }
	 filterDB.insert(filter_map_node(name, fdtemp));
      }
   }

   //-----------------------------------  Define processed channels.
   if (jstak.push_member("processed-channels")) {
      jsonStack::autopop pop_proc_chans(jstak);
      if (Debug()) cout << "config: loop over filters" << endl;
      for (int i=0; jstak.iterate(); i++) {
	 jsonStack::autopop pop_iter(jstak);
	 string name;
	 if (!jstak.fetch_data("name", name)) {
	    cerr << "process-channel must specify a name" << endl;
	    syntax = true;
	    continue;
	 }
	 process_channel chan(name);
	 string mode("copy"), x_input(name), y_input, filter;
	 if (jstak.fetch_data("x-input", x_input)) chan.set_channel_x(x_input);
	 if (jstak.fetch_data("y-input", y_input)) chan.set_channel_y(y_input);
	 if (jstak.fetch_data("filter", filter)) {
	    auto filtp = filterDB.find(filter);
	    if (filtp == filterDB.end()) {
	       cerr << "Filter: " << filter << " is not defined." << endl;
	       syntax = true;
	       continue;
	    }
	    chan.set_filter(filtp->second);
	 }
	 if (jstak.fetch_data("mode", mode)) chan.set_process_type(mode);
	 else                                chan.set_process_type("write");
	 proc_list.push_back(chan);
      }
   }
}

//======================================  Align write buffer start times
void
DQ_Module::align_buffer(const Time& t) {
   for (auto itr=proc_list.begin(); itr != proc_list.end(); itr++) {
      TSeries& tsi(itr->buffer());
      if (tsi.empty()) continue;
      if (tsi.getStartTime() < t) {
	 tsi.eraseStart(t - tsi.getStartTime());
      }
   }
}

//======================================  Request all input channels.
void
DQ_Module::book_channels(void) {
   if (Debug()) cout << "DQ_Module::book_channels: Booking channels" << endl;

   //--------------------------------- combine all channels
   DQ_bit::chanlist_type combined;
   size_t nFlags = flag_vect.size();
   if (nFlags) {
      combined = flag_vect[0]->channel_list();
      for (size_t i=1; i<nFlags; i++) {
	 const DQ_bit::chanlist_type& chanlist = flag_vect[i]->channel_list();
	 combined.insert(combined.end(), chanlist.begin(), chanlist.end());
      }
   }

   //---------------------------------  add in processed channel inputs
   for (auto itr=proc_list.begin(); itr!= proc_list.end(); itr++) {
      string chanX = itr->channel_x();
      if (chan_type(chanX) == 0) {
	 combined.push_back(chanX);
	 size_t inx = chanX.find("/");
	 if (inx != string::npos) {
	    chanX.erase(0, inx+1);
	    itr->set_channel_x(chanX);
	 }
      }
      string chanY = itr->channel_y();
      if (!chanY.empty() && chan_type(chanY) == 0) {
	 combined.push_back(chanY);
	 size_t inx = chanY.find("/");
	 if (inx != string::npos) {
	    chanY.erase(0, inx+1);
	    itr->set_channel_y(chanY);
	 }
      }
   }
   if (combined.empty()) return;

   //---------------------------------- make a unique list.
   std::sort(combined.begin(), combined.end());
   size_t nChans = combined.size();
   size_t j = 0;
   for (size_t i=1; i<nChans; i++) {
      if (combined[i] != combined[j]) {
	 j++;
	 if (i != j) combined[j] = combined[i];
      }
   }
   j++;
   combined.erase(combined.begin() + j, combined.end());

   //-----------------------------------  Book all unique channels
   nChans = combined.size();
   for (size_t i=0; i<nChans; i++) {
      getDacc().addChannel(combined[i]);
      if (Debug()) cout << "DQ_Module::book_channels: Added channel "
			<< combined[i] << endl;
   }
}

//======================================  Channel type finder
int 
DQ_Module::chan_type(const std::string& chan) const {
   for (auto itr=proc_list.begin(); itr != proc_list.end(); itr++) {
      if (itr->name() != chan) continue;
      if (itr->channel_x() == chan || itr->channel_y() == chan) continue;
      if (Debug() > 1) cout << "chan_type: channel; " << chan << " type: "
			    << 1 << endl;
      return 1;
   }

   if (flag_map.find(chan) != flag_map.end()) {
      if (Debug() > 1) cout << "chan_type: channel; " << chan << " type: "
			    << 2 << endl;
      return 2;
   }

   if (Debug() > 1) cout << "chan_type: channel: " << chan << " type: "
			 << 0 << endl;
   return 0;
}

//======================================  Channel series finder
const TSeries&
DQ_Module::find(const std::string& chan) const {

   //-----------------------------------  Look for a processed channel
   for (auto itr=proc_list.begin(); itr != proc_list.end(); itr++) {
      if (itr->name() != chan) continue;
      if (itr->channel_x() == chan || itr->channel_y() == chan) continue;
      return itr->ref_series();
   }

   //-----------------------------------  look for a flag vector channel
   auto p = flag_map.find(chan);
   if (p != flag_map.end()) return p->second.ref_series();

   //-----------------------------------  Now try a raw input channel
   const TSeries* ts = getDacc().refData(chan);
   if (ts) return *ts;

   if (Debug() > 1) cout << "find: Could not find channel: " << chan << endl;
   return null_ts;
}

//======================================  calculate the flag bit
void
DQ_Module::set_flag(DQ_bit* dq_def, const Time& t0, Interval dT) {
   //-----------------------------------  Get the input channel list
   DQ_bit::tser_vect data;
   const DQ_bit::chanlist_type& chanlist = dq_def->channel_list();
   size_t nChan = chanlist.size();
   data.reserve(nChan);
   for (size_t i=0; i<nChan; i++) {
      const TSeries* tsi = getDacc().refData(chanlist[i]);
      if (tsi) {
	 data.push_back(tsi->extract(t0, dT));
      } 
      else if (flag_map.find(chanlist[i]) != flag_map.end()) {
	 data.push_back(flag_map[chanlist[i]].ref_series().extract(t0,dT));
      }
      else {
	 ostringstream msg;
	 msg << "DQ_Module: Unable to find channel " << chanlist[i] 
             << " to evaluate " << dq_def->dq_name() << endl;
	 throw std::runtime_error(msg.str());
      }
   }

   //-----------------------------------  Calculate the bit value
   bool bit = dq_def->bit_value(data);
   if (dq_def->invert()) bit = !bit;

   //-----------------------------------  And in the flag content
   if (!bit) {
      flag_map_iter out_chan = flag_map.find(dq_def->bit_channel());
      TSeries& tsout = out_chan->second.ref_series();
      size_t inx = tsout.getBin(t0);
      DVectU* dv = dynamic_cast<DVectU*>(tsout.refDVect());
      if (!dv) {
	 cerr << "Unable to find data vector for " << dq_def->bit_channel()
	      << endl;
      } else {
	 (*dv)[inx] &= ~(1 << dq_def->bit_number());
      }
   }
}

//======================================  Initialize defined flag instances.
void
DQ_Module::init_all(void) {
   size_t nFlags = flag_vect.size();
   for (size_t i=0; i<nFlags; i++) {
      flag_vect[i]->init();
   }
}

//======================================  Define output channels.
void
DQ_Module::init_output_channels(void) {
   size_t nFlags = flag_vect.size();
   for (size_t iflag=0; iflag < nFlags; ++iflag) {
      string chan = flag_vect[iflag]->bit_channel();
      flag_map_iter itr = flag_map.find(chan);
      if (itr == flag_map.end()) {
	 dq_bitmask temp;
	 flag_map.insert(flag_map_type::value_type(chan, temp));
	 itr = flag_map.find(chan);
	 if (itr == flag_map.end()) throw runtime_error("unableto add channel");
	 process_channel pr(chan);
	 pr.set_process_type(process_channel::pk_copy);
      }
      itr->second.enable(flag_vect[iflag]->bit_number());
   }
}

//======================================  Process a stride.
void 
DQ_Module::ProcessData(void) {
   uint32_t data=0;

   if (!current || current < getDacc().getFillTime()) {
      flush(); // noop if !current
      reset();
      current = getDacc().getFillTime();
      write_time = current;
   }
   
   //-----------------------------------  Get time structure for a bit channel
   Interval dT = getDacc().getFillStride();
   Interval fstep(flag_step);
   int nstep = int(double(dT) / fstep);

   //-----------------------------------  Initialize output bit masks.
   for (flag_map_iter i=flag_map.begin(); i != flag_map.end(); i++) {
      i->second.preset(current, fstep, nstep);
      i->second.ref_series().setName((i->first + " (map)").c_str());
   }
   
   //-----------------------------------  Step over times
   size_t nFlags = flag_vect.size();
   for (int istep=0; istep < nstep; istep++) {
      Time t0 = current + fstep * double(istep);
      //---------------------------------  Evaluate each flag
      for (size_t iflag=0; iflag<nFlags; iflag++) {	 
	 set_flag(flag_vect[iflag], t0, Interval(flag_step));
      }
   }

   //-----------------------------------  Set the parity bit and invert data
   if (parity_bit >= 0 || invert) {
      const uint32_t parity_mask = (parity_bit < 0) ? 0 : (1 << parity_bit);
      for (flag_map_iter out_chan=flag_map.begin(); 
	   out_chan != flag_map.end();
	   out_chan++) { 
         TSeries& tsout = out_chan->second.ref_series();
         DVectU* dv = dynamic_cast<DVectU*>(tsout.refDVect());
	 size_t N=tsout.getNSample();
	 for (size_t i=0; i<N; i++) {
	    uint32_t dvi = (*dv)[i];
	    if (parity_mask) {
	       dvi &= ~parity_mask;
	       if (!odd_parity(dvi)) dvi |= parity_mask;
            }
	    if (invert) dvi = ~dvi;
	    (*dv)[i] = dvi;
	 }
      } // loop over DQ vectors
   }

   //----------------------------------- loop over processed channels
   for (auto itr=proc_list.begin(); itr!=proc_list.end(); itr++) {
      const TSeries& X(find(itr->channel_x()));
      if (itr->channel_y().empty()) {
	 itr->apply(X);
      } else {
	 const TSeries& Y(find(itr->channel_y()));
	 itr->apply(X, Y);
      }
   }

   current += dT;
   
   //-----------------------------------  Clear copied output bit masks.
   for (auto i=flag_map.begin(); i != flag_map.end(); i++) {
      TSeries& ts(i->second.ref_series());
      ts.eraseStart(current - ts.getStartTime());
   }
   
   //-----------------------------------  Buffer for writing
   Time t0(0), tEnd(0);
   for (auto itr=proc_list.begin(); itr != proc_list.end(); itr++) {
      const TSeries& tsi = itr->buffer();
      if (!itr->write()) continue;
      if (tsi.empty()) {
	 cerr << "write buffer for " << itr->name() << " is empty" << endl;
	 continue;
      }
      if (!t0) {
	 t0   = tsi.getStartTime();
	 tEnd = tsi.getEndTime();
	 continue;
      }
      if (t0   < tsi.getStartTime())  t0 = tsi.getStartTime();
      if (tEnd > tsi.getEndTime())  tEnd = tsi.getEndTime();
   }
   if (Debug()) cout << "Available data: " << t0 << " - " << tEnd << endl;
   
   if (t0 < write_time) align_buffer(write_time);
   
   //-----------------------------------  Write a frame
   while (write_time + dT <= tEnd) {
      //cout << "write_frame(" << write_time << ", " << dT << ")" << endl;
      write_frame(write_time, dT);
      write_time += dT;
   }

   align_buffer(write_time);
}

//======================================  Reset filters and DQ flag channels
void
DQ_Module::reset(void) {
   init_all();
   for (auto itr=proc_list.begin(); itr != proc_list.end(); itr++) {
      itr->reset();
   }
   current    = Time(0);
   write_time = current;
}

//======================================  Write one frame
void
DQ_Module::write_frame(const Time& t0, Interval dT) {
   fr_writer.buildFrame(t0, dT);

   //-----------------------------------  Copy requested history records.
   size_t nHist = copy_history.size();
   for (size_t i=0; i<nHist; i++) {
      string frametype = copy_history[i];
      int inx = getDacc().frame_type(frametype);
      if (inx < 0) {
	 if (frametype[0] >= '0' && frametype[0] <= '9') {
	    inx = strtoul(frametype.c_str(), 0, 0);
	 } else {
	    continue;
         }
      }
      if (inx < 0 || inx > getNStream()) {
	 static bool first_er1=true;
	 if (first_er1) {
	    cerr << "DQ_Module::write_frame: " << frametype
		 << " is not a valid stream identifier." << endl;
	    first_er1 = false;
	 }
      }
      else if (!getDacc(inx) || !getDacc(inx)->getFrame()) {
	 static bool first_er2=true;
	 if (first_er2) {
	    cerr << "DQ_Module::write_frame: No history records available "
		 << "for stream " << frametype << endl;
	    first_er2 = false;
	 }
      }
      else {
	 fr_writer.copyHistory(*(getDacc(inx)->getFrame()));
      }
   }

   //-----------------------------------  Add a history record
   ostringstream config;
   config << "Configuration file: " <<  config_file << ", version: "
	  << config_version << ", crc32: " << config_crc32;
   fr_writer.addHistory("dq-module", Now(), config.str());

   //-----------------------------------  write requested channels.
   for (auto itr=proc_list.begin(); itr != proc_list.end(); itr++) {
      if (itr->write()) {
	 string tsname = itr->name();
	 const TSeries& tsi = itr->buffer();
	 if (tsi.empty()) {
	    cerr << "Error: Processed channel: " << tsname
		 << " is not available." << endl;
	 } else {
	    fr_writer.addProcSeries(tsname, tsi.extract(t0, dT));
	 }
      }
   }

   //-----------------------------------  Write DQ flag series
   //  DQ flags entered as copy channels.
   //for (auto i=flag_map.begin(); i != flag_map.end(); i++) {
   //   fr_writer.addProcSeries(i->first, i->second.ref_series());
   //}

   //-----------------------------------  Open a writer to a frame file
   string outdir(frame_path.dir_name(t0));
   if (!frame_path.online()) {
      frame_path.make_dir(outdir);
      string file_path = frame_path.file_path(t0, dT);
      if (Debug()) cout << "  opening file: " << file_path << endl;
      if (fr_writer.open(file_path.c_str(), true)) {
	 fr_writer.erase();
	 cerr << "DQ_Module: Failed to open frame " << file_path
	      << " for output." << endl;
	 return;
      }
   }
   
   //----------------------------------  Open a writer to shared memory.
   else {
      if (fr_writer.open(outdir.c_str(), true)) {
	 fr_writer.erase();
	 cerr << "DQ_Module: Failed to open partition " << outdir
	      << " for output." << endl;
	 return;
      }
   }
 
   //-----------------------------------  Write the frame
   try {
      if (fr_writer.writeFrame()) {
	 cerr << "DQ_Module: Error writing frame with GPS " << t0.getS()
	      << " to output Directory " << outdir << endl;
      } else {
	 frames_out++;
      }
   } catch (std::exception& e) {
      cerr << "Caught exception in writeFrame: " << e.what() << endl;
   }
   
   //-----------------------------------  Clean up
   fr_writer.close();
   fr_writer.erase();
}

//======================================  DQ channel constructor
DQ_Module::dq_bitmask::dq_bitmask(void) 
  : _enable_mask(0)
{
}

//======================================  DQ channel destructor
DQ_Module::dq_bitmask::~dq_bitmask(void) {
}

//======================================  Set an output channel for one stride
void
DQ_Module::dq_bitmask::preset(const Time& t0, Interval dT, size_t N) {
   DVectU data(N);
   for (size_t i=0; i<N; i++) data[i] = _enable_mask;
   _series.setData(t0, dT, data.clone());
}

//======================================  Filter template default constructor
DQ_Module::filter_template::filter_template(void)
   : _type(kf_none)
{
}

//======================================  Filter template initial constructor
DQ_Module::filter_template::filter_template(const std::string& name,
					    const FilterBase& base,
					    filter_enum type) {
   _name = name;
   _filter.reset(base.clone());
   _type = type;
}

//======================================  Filter template copy constructor
DQ_Module::filter_template::filter_template(const filter_template& ft) {
   *this = ft;
}

//======================================  Filter template assignment operator
DQ_Module::filter_template& 
DQ_Module::filter_template::operator=(const filter_template& ft) {
   _name = ft._name;
   if (ft._filter) _filter.reset(ft._filter->clone());
   _type = ft._type;
   return *this;
}

//======================================  Filter template destructor
DQ_Module::filter_template::~filter_template(void) {
}
   
//======================================  Filter template apply pipe method
TSeries
DQ_Module::filter_template::apply(const TSeries& x) {
   if (_type == kf_none) return x;
   if (_type != kf_pipe) {
      throw logic_error("filter_template::apply: Invalid filter type");
   }
   return dynamic_cast<Pipe*>(_filter.get())->apply(x);
}
   
//======================================  Filter template apply ypipe method
TSeries
DQ_Module::filter_template::apply(const TSeries& x, const TSeries& y) {
   if (_type != kf_ypipe) {
      throw logic_error("filter_template::apply: Invalid filter type");
   }
   return dynamic_cast<YPipe*>(_filter.get())->apply(x, y);
}

void
DQ_Module::filter_template::reset(void) {
   if (_type != kf_none) _filter->reset();
}

std::string 
DQ_Module::filter_template::type(void) const {
   switch (_type) {
   case kf_none:
      return "none";
   case kf_pipe:
      return "pipe";
   case kf_ypipe:
      return "ypipe";
   default:
      break;
   }
   return "undefined";
}

//=======================================  Processed channel class

//=======================================  Processed channel constructor
DQ_Module::process_channel::process_channel(const std::string& chan) {
    _process_type = pk_undefined;
    _chan_name = chan;
    _chan_inx = chan;
}
    
//=======================================  Processed channel destructor
DQ_Module::process_channel::~process_channel(void) {
}
   
//======================================  Filter template apply pipe method
void
DQ_Module::process_channel::apply(const TSeries& x) {
   _output = _filt.apply(x);
   _output.setName((_chan_name + "(out)").c_str());
}
   
//======================================  Filter template apply ypipe method
void
DQ_Module::process_channel::apply(const TSeries& x, const TSeries& y) {
   _output = _filt.apply(x, y);
   _output.setName((_chan_name + "(out)").c_str());
}

//=======================================  Buffer data for output.
TSeries& 
DQ_Module::process_channel::buffer(void) {
   if (write() && !ref_series().empty()) {
      if (_write_buffer.Append(_output)) {
	 throw runtime_error("process_channel: write buffer not in synch");
      }
   }
   ref_series().Clear();
   _write_buffer.setName(_chan_name.c_str());
   return _write_buffer;
}

//=======================================  Write the process channel status
void
DQ_Module::process_channel::dump(std::ostream& out) const {
   out << " name:           " << name() << endl;
   out << " type:           " << type() << endl;
   if (filter().type() != "none") {
      out << " filter-name:    " << filter().name() << endl;
      out << " filter-type:    " << filter().type() << endl;
   }
   out << " x-input:        " << channel_x() << endl;
   if (!channel_y().empty()) {
      out << " y-input:        " << channel_y() << endl;
   }
}

//=======================================  Flush data from GateGen filters
void
DQ_Module::process_channel::flush(void) {
   GateGen* gg = dynamic_cast<GateGen*>(_filt.get());
   if (gg) _output.Append(gg->flush());
}

//=======================================  Reset filter
void
DQ_Module::process_channel::reset(void) {
   _filt.reset();
   _write_buffer.Clear();
}

//=======================================  Set X channel name
void
DQ_Module::process_channel::set_channel_x(const std::string& xname) {
   _chan_inx = xname;
}

//========================================  Set Y channel name
void
DQ_Module::process_channel::set_channel_y(const std::string& yname) {
   _chan_iny = yname;
}

//========================================  Set the filter
void
DQ_Module::process_channel::set_filter(const filter_template& filt) {
   _filt = filt;
}

//========================================  Set the process type
void
DQ_Module::process_channel::set_process_type(process_enum mode) {
   _process_type = mode;
}

void
DQ_Module::process_channel::set_process_type(const std::string& mode) {
      if (mode == "none") set_process_type(pk_none);
      else if (mode == "nowrite") set_process_type(pk_nowrite);
      else if (mode == "copy") set_process_type(pk_copy);
      else if (mode == "write") set_process_type(pk_write);
      else set_process_type(pk_undefined);
}

std::string 
DQ_Module::process_channel::type(void) const {
   switch (_process_type) {
   case pk_undefined:
      return "undefined";
   case pk_none:
      return "none";
   case pk_nowrite:
      return "nowrite";
   case pk_copy:
      return "copy";
   case pk_write:
      return "write";
   }
   return "undefined";
}
